# org@hardwaremassagecircus.com
#
# todo:
# ACK system
# functions for server commands
# use sysread
# register warnings
# timer
# perl Makefile.PL stuff

package Kismet;
require Exporter;
use strict;
use warnings;
use Socket qw(:DEFAULT :crlf);
use Carp;

our @ISA = qw(Exporter);
our @EXPORT = qw(KismetRun);
our @EXPORT_OK = qw(KismetInit KismetLoop);
our %EXPORT_TAGS = (manual => [qw(KismetInit KismitLoop)], 
			callback => [qw(KismetRun)]);
our $VERSION = 1.00;

our $TIMEOUT = 0.5;

sub new {
	my ($class, $host, $port) = @_;
	my $self = {'_host', $host, '_port', $port};
	my $ip = inet_aton($host)
		or croak("Cannot resolve host name: $!\n");
	$self->{'_addr'} = sockaddr_in($port, $ip);
	socket($self->{'_socket'}, PF_INET, SOCK_STREAM, getprotobyname('tcp'))
		or croak("Cannot get a socket: $!\n");
	select((select($self->{'_socket'}), $| = 1)[0]);
	$self->{'_startup'} = 1;
	return bless $self, $class;
}

sub DESTROY {
	my $self = shift;
	shutdown $self->{'_socket'}, 2;
}

sub prestartup {
	my $self = shift;
	connect($self->{'_socket'}, $self->{'_addr'})
		or croak("Cannot connect to kismet server $self->{'_host'}:$self->{'_port'}: $!\n");
}

sub poststartup {
	my $self = shift;
	my $rin = my $rout = '';
	vec ($rin, fileno($self->{'_socket'}), 1) = 1;
	while (my $pr = shift @{$self->{'_precallback'}}) {
		$self->register(@$pr);
	}
	while (my $bl = shift @{$self->{'_backlog'}}) {
		$self->dispatch(@$bl);
	}
}

sub KismetRun {
	my @objects = @_;
	KismetInit(@objects);
	while(defined KismetLoop(@objects)) {}
}

sub KismetInit {
	my @objects = @_;
	$_-> prestartup() foreach @objects;
}

sub KismetLoop {
	my @objects = @_;

	my $rin = '';
	my $alive = 0;
	foreach (@objects) {
		if (!$_->{'_shutdown'}) {
			vec ($rin, fileno($_->{'_socket'}) , 1) = 1;
			$alive++;
		}
	}
	return undef unless $alive;

	my $ready = 0;
	if ($ready = select(my $rout = $rin, undef, undef, $TIMEOUT)) {
		foreach (@objects) {
			if (vec $rout, fileno($_->{'_socket'}), 1) {
				$_->readsocket();
			}
		}
	}
	return $ready;
}

sub readsocket {
	my $self = shift;
	my $rin = '';
	vec ($rin, fileno($self->{'_socket'}), 1) = 1;
	{
	my $soc = $self->{'_socket'};
	if (eof($soc)) {
		$self->{'_shutdown'} = 1;
		return;
	}
	my $x = <$soc>;
	my ($p, $a) = parse(\$x);
	if (defined $p) {
		if ($self->{'_startup'}) {
			if ($p eq "PROTOCOLS") {
				$self->{'_startup'} = 0;
				foreach (split /,/, $a->[0]) {
					print $soc "!0 CAPABILITY $_$CRLF";
					$self->{'_startup'}++;
				}
			}
			elsif ($p eq "CAPABILITY") {
				my $x = 0;
				my $uline = "";
				foreach (split /,/, $a->[1]) {
					$self->{'_capmap'}->{$a->[0]}->{$_} = $x++;
					$uline = $_;
				}
				if ($x == 1) {
					$self->{'_capmap'}->{$a->[0]}->{$uline} = '*';
				}
				$self->{'_startup'}--;
			}
			push @{$self->{'_backlog'}}, [$p, $a];
			unless ($self->{'_startup'}) {
				$self->poststartup();
			}
		} else {
			$self->dispatch($p, $a);
		}
	}
	redo if select(my $rout = $rin, undef, undef, 0);
	}
}

sub preregister {
	my ($self, $proto, $callback, $opts) = @_;
	$opts ||= '*';
	if (!defined($proto)) {
		croak("No protocall given to register callback");
	}
	if (ref($callback) ne "CODE") {
		croak("Attempted to register no/invalid callback for $proto\n");
	}
	push @{$self->{'_precallback'}}, [$proto, $callback, $opts];
}

sub register {
	my ($self, $proto, $callback, $opts) = @_;
	$opts ||= '*';
	if ($self->{'_startup'}) {
		$self->preregister($proto, $callback, $opts);
		return;
	}
	if (!defined($proto)) {
		croak("No protocall given to register callback");
	}
	if (ref($callback) ne "CODE") {
		croak("Attempted to register no/invalid callback for $proto\n");
	}
	$proto = uc $proto;
	if (exists $self->{'_capmap'}->{$proto}) {
		my @o = ();
		if ($opts eq '*') {
			@o = keys %{$self->{'_capmap'}->{$proto}};
		} else {
			$opts = lc $opts;
			foreach (split /,/, $opts) {
				if (exists $self->{'_capmap'}->{$proto}->{$_}) {
					push @o, $_;
				}
			}
		}
		$self->{'_callback'}->{$proto} = [$callback, \@o];
		print {$self->{'_socket'}} "!0 ENABLE $proto *$CRLF";
	} else {
		carp("$proto not a valid protocall for this server\n");
	}
}

sub dispatch {
	my ($self, $proto, $args) = @_;
	if (exists $self->{'_callback'}->{$proto}) {
		my %hash = ();
		foreach (@{$self->{'_callback'}->{$proto}->[1]}) {
			my $id = $self->{'_capmap'}->{$proto}->{$_};
			if ($id eq '*') {
				$hash{$_} = join " ", @$args;
			} else {
				$hash{$_} = $args->[$id];
			}
		}
		$self->{'_callback'}->{$proto}->[0](\%hash);
	}
}

sub parse {
	my $raw = shift;
	my $proto = '';
	my @args = ();

###NOT FINISHED
#	if ($$raw !~ /^\*/) {
#		my $ind = index $$raw, "\n*";
#		if ($ind == -1) {
#			$$raw = '';
#			return undef;
#		}
#		substr $$raw, 0, $ind+1, '';
#	}
#*goobiei\001 \001dafe\n*rebok\001\n*TEST: bla\n
###END

	my $pos = my $end = 0;
	my $string = '';

	{
	$end = index $$raw, "\n", $pos;
	if ($end == -1) {
		return undef;
	}
	$string = substr $$raw, 0, $end+1;
	my $c = $string =~ tr/\001//;
	if ($c % 2) {
		$pos = $end+1;
		redo;
	}
	}

	substr $$raw, 0, $end+1, '';

	if ($string =~ s/^\*([A-Z]*): //) {
		chomp $string;
		$proto = $1;
		my @x = split / /, $string;
		for (my $i=0; $i <= $#x; $i++) {
			if ($x[$i] =~ /^\001/) {
				my $bits = $x[$i];
				$bits .= " " . $x[++$i]	while length($bits) == 1 || $bits !~ /\001$/;
				push @args, substr($bits, 1, -1);
			} else {
				push @args, $x[$i];
			}
		}
		return $proto, \@args;		
	}
	return undef;	
}

1;
